"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2020 Deepkit UG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.JSONObjectCollection = exports.Collection = exports.CollectionPagination = void 0;
/**
 * This is a collection object that contains items of an certain entity.
 * This collection "lives" in the sense that its items are automatically
 * updated, added and removed. When such a change happens, an event is triggered* you can listen on.
 */
const rxjs_1 = require("rxjs");
const core_rxjs_1 = require("@deepkit/core-rxjs");
const core_1 = require("@deepkit/core");
class CollectionPagination {
    constructor(collection) {
        this.collection = collection;
        this.event = new rxjs_1.Subject();
        this._parameters = {};
        this.page = 1;
        this.total = 0;
        this.itemsPerPage = 50;
        this.sort = [];
        this.active = false;
    }
    setParameters(values) {
        this._parameters = values || {};
        return this;
    }
    resetParameters() {
        this._parameters = {};
        return this;
    }
    setParameter(name, value) {
        this._parameters[name] = value;
        return this;
    }
    getParameter(name) {
        return this._parameters[name];
    }
    getParameters() {
        return this._parameters;
    }
    getHash() {
        return JSON.stringify([
            this.page, this.itemsPerPage, this.sort, this._parameters
        ]);
    }
    getPagingHash() {
        return JSON.stringify([
            this.page, this.itemsPerPage, this.sort
        ]);
    }
    getParametersHash() {
        return JSON.stringify([
            this._parameters
        ]);
    }
    setPage(page) {
        this.page = page;
        return this;
    }
    hasSort() {
        return this.sort.length > 0;
    }
    getSort() {
        return this.sort;
    }
    setSort(order) {
        this.sort = order;
        return this;
    }
    orderByField(field, direction = 'asc') {
        this.sort = [{ field: field, direction: direction }];
        return this;
    }
    /**
     * Sends current pagination setting to the server and refreshes the collection if necessary.
     */
    apply() {
        this.applyPromise = new Promise((resolve, reject) => {
            this.applyPromiseResolver = resolve;
            this.event.next({ type: 'apply' });
        });
        return this.applyPromise;
    }
    /**
     * Triggered from the server when the apply() finished. Doesn't matter if we got actual updates or not. It's important
     * to distinguish because collection.nextStateChange is not reliable enough
     * (as entity updates could happen between apply and applyFinish, which would trigger nextStateChange).
     *
     * @private
     */
    _applyFinished() {
        if (this.applyPromiseResolver) {
            this.applyPromiseResolver();
            delete this.applyPromise;
            delete this.applyPromiseResolver;
        }
    }
    setTotal(total) {
        this.total = total;
        return this;
    }
    setItemsPerPage(items) {
        this.itemsPerPage = items;
        return this;
    }
    isActive() {
        return this.active;
    }
    /**
     * It's not possible to activate this later on. This is set on the server side only.
     */
    _activate() {
        this.active = true;
        return this;
    }
    getPage() {
        return this.page;
    }
    getItemsPerPage() {
        return this.itemsPerPage;
    }
    getTotal() {
        return this.total;
    }
    getPages() {
        return Math.ceil(this.getTotal() / this.getItemsPerPage());
    }
    isPageValid() {
        return this.page > this.getPages();
    }
}
exports.CollectionPagination = CollectionPagination;
class Collection extends rxjs_1.ReplaySubject {
    constructor(classType) {
        super(1);
        this.classType = classType;
        this.event = new rxjs_1.Subject;
        this.removed = new rxjs_1.Subject();
        this.added = new rxjs_1.Subject();
        this.teardowns = [];
        this.items = [];
        this.itemsMapped = {};
        this.deepChange = new rxjs_1.Subject();
        this.pagination = new CollectionPagination(this);
        this.batchActive = false;
        this.batchNeedLoaded = false;
        this.entitySubjects = {};
    }
    getEntitySubject(idOrItem) {
        const id = idOrItem instanceof this.classType ? idOrItem.id : String(idOrItem);
        return this.entitySubjects[id];
    }
    has(id) {
        return 'undefined' !== typeof this.itemsMapped[id];
    }
    get(id) {
        return this.itemsMapped[id];
    }
    batchStart() {
        this.batchActive = true;
        this.event.next({ type: 'batch/start' });
    }
    batchEnd() {
        this.batchActive = false;
        if (this.batchNeedLoaded) {
            this.loaded();
        }
        this.batchNeedLoaded = false;
        this.event.next({ type: 'batch/end' });
    }
    /**
     * Resolves when next change happened.
     */
    get nextStateChange() {
        if (!this.nextChange) {
            this.nextChange = new rxjs_1.Subject();
        }
        return this.nextChange.toPromise();
    }
    /**
     * Unsubscribe from the backend stream.
     */
    async unsubscribe() {
        await super.unsubscribe();
        this.pagination.event.unsubscribe();
        for (const teardown of this.teardowns) {
            await core_rxjs_1.tearDown(teardown);
        }
        this.teardowns.splice(0, this.teardowns.length);
    }
    addTeardown(teardown) {
        this.teardowns.push(teardown);
    }
    index(item) {
        return this.items.indexOf(item);
    }
    /**
     * Returns the page zero-based of the current item.
     */
    getPageOf(item, itemsPerPage = 10) {
        const index = this.index(item);
        if (-1 === index)
            return 0;
        return Math.floor(index / itemsPerPage);
    }
    reset() {
        this.items = [];
        this.itemsMapped = {};
    }
    all() {
        return this.items;
    }
    count() {
        return this.items.length;
    }
    ids() {
        const ids = [];
        for (const i of this.items) {
            ids.push(i.id);
        }
        return ids;
    }
    empty() {
        return 0 === this.items.length;
    }
    map() {
        return this.itemsMapped;
    }
    loaded() {
        if (this.batchActive) {
            this.batchNeedLoaded = true;
            return;
        }
        this.batchNeedLoaded = false;
        if (this.isStopped) {
            throw new Error('Collection already unsubscribed');
        }
        this.next(this.items);
        if (this.nextChange) {
            this.nextChange.complete();
            delete this.nextChange;
        }
    }
    seItem(id, item) {
        if (this.itemsMapped[item.id]) {
            const index = this.items.indexOf(this.itemsMapped[item.id]);
            this.items[index] = item;
            this.itemsMapped[item.id] = item;
        }
        else {
            this.items.push(item);
            this.itemsMapped[item.id] = item;
        }
    }
    set(items, withEvent = true) {
        for (const item of items) {
            if (!this.itemsMapped[item.id]) {
                this.added.next(item);
            }
            delete this.itemsMapped[item.id];
        }
        for (const deleted of core_1.each(this.itemsMapped)) {
            this.removed.next(deleted);
        }
        this.itemsMapped = {};
        this.items = items;
        for (const item of items) {
            this.itemsMapped[item.id] = item;
        }
        if (withEvent) {
            this.event.next({ type: 'set', items: items });
            this.loaded();
        }
    }
    setSort(ids) {
        this.items.splice(0, this.items.length);
        for (const id of ids) {
            this.items.push(this.itemsMapped[id]);
        }
        this.event.next({ type: 'sort', ids: ids });
        this.loaded();
    }
    removeMany(ids, withEvent = true) {
        for (const id of ids) {
            const item = this.itemsMapped[id];
            delete this.itemsMapped[id];
            const index = this.items.indexOf(item);
            if (-1 !== index) {
                this.removed.next(this.items[index]);
                this.items.splice(index, 1);
            }
        }
        if (withEvent) {
            this.event.next({ type: 'removeMany', ids: ids });
            this.loaded();
        }
    }
    add(item, withEvent = true) {
        if (!item) {
            throw new Error(`Trying to insert a ${core_1.getClassName(this.classType)} collection item without value`);
        }
        this.added.next(item);
        if (this.itemsMapped[item.id]) {
            const index = this.items.indexOf(this.itemsMapped[item.id]);
            this.items[index] = item;
            this.itemsMapped[item.id] = item;
        }
        else {
            this.items.push(item);
            this.itemsMapped[item.id] = item;
        }
        if (withEvent) {
            this.event.next({ type: 'add', item: item });
            this.loaded();
        }
    }
    remove(id, withEvent = true) {
        if (this.itemsMapped[id]) {
            const item = this.itemsMapped[id];
            delete this.itemsMapped[id];
            const index = this.items.indexOf(item);
            if (-1 !== index) {
                this.items.splice(index, 1);
            }
            if (withEvent) {
                this.event.next({ type: 'remove', id: item.id });
                this.removed.next(item);
                this.loaded();
            }
        }
    }
}
exports.Collection = Collection;
class JSONObjectCollection extends Collection {
}
exports.JSONObjectCollection = JSONObjectCollection;
//# sourceMappingURL=collection.js.map