"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2020 Deepkit UG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.observeItem = exports.ItemObserver = void 0;
const type_1 = require("@deepkit/type");
const core_1 = require("@deepkit/core");
const fast_json_patch_1 = require("fast-json-patch");
const dot_prop_1 = require("dot-prop");
const rxjs_1 = require("rxjs");
const core_2 = require("./core");
class ItemObserver extends rxjs_1.BehaviorSubject {
    constructor(value) {
        super(undefined);
        this.ignore = ['version'];
        if (value) {
            this.start(value);
        }
    }
    static fromValueOrSubject(valueOrSubject, ignore = []) {
        const observer = new ItemObserver(undefined);
        observer.ignore.push(...ignore);
        observer.start(valueOrSubject);
        return observer;
    }
    /**
     * After creating an ItemObserver, the user requires to operate on this object. Otherwise changes are not detected.
     */
    get snapshot() {
        if (!this._snapshot) {
            throw new Error('Observer not started');
        }
        return this._snapshot;
    }
    start(valueOrSubject) {
        if (!valueOrSubject) {
            throw new Error('No value given for observer');
        }
        this.valueOrSubject = valueOrSubject;
        if (this.valueOrSubject instanceof rxjs_1.BehaviorSubject) {
            //this should not produce memory leaks, as the EntitySubject should always be unsubscribed when done,
            // so this Subscription will be automatically removed as well.
            this._snapshot = type_1.cloneClass(this.valueOrSubject.getValue());
            if (this.valueOrSubject instanceof core_2.EntitySubject) {
                this.valueOrSubject.patches.subscribe((patches) => {
                    if (this.valueOrSubject instanceof core_2.EntitySubject) {
                        if (this.old) {
                            const plainPatches = type_1.jsonSerializer.for(Object.getPrototypeOf(this.valueOrSubject.value).constructor).partialSerialize(patches);
                            for (const [i, v] of core_1.eachPair(plainPatches)) {
                                dot_prop_1.set(this.old, i, v);
                            }
                        }
                        for (const [i, v] of core_1.eachPair(patches)) {
                            dot_prop_1.set(this._snapshot, i, v);
                        }
                        this.next(this._snapshot);
                    }
                });
            }
        }
        else {
            this._snapshot = type_1.cloneClass(this.valueOrSubject);
        }
        this.old = this.createState();
        this.next(this._snapshot);
    }
    get original() {
        if (this.valueOrSubject) {
            return this.valueOrSubject instanceof core_2.EntitySubject ? this.valueOrSubject.getValue() : this.valueOrSubject;
        }
        return;
    }
    createState() {
        if (this._snapshot && this.original) {
            const item = type_1.jsonSerializer.for(type_1.getClassSchema(this.original)).serialize(this._snapshot);
            for (const p of this.ignore) {
                delete item[p];
            }
            return item;
        }
        else {
            return {};
        }
    }
    changed() {
        if (this.old) {
            const newState = this.createState();
            const ops = fast_json_patch_1.compare(this.old, newState);
            return ops.length !== 0;
        }
        return false;
    }
    /**
     * Returns the plain patch values, key value.
     */
    normalizePatch(ops) {
        if (!this._snapshot)
            return undefined;
        const patches = {};
        const doc = this._snapshot;
        for (const op of ops) {
            const path = op.path.slice(1).replace(/\//g, '.');
            if (op.op === 'add' || op.op === 'replace') {
                patches[path] = core_1.getPathValue(doc, path);
            }
            else if (op.op === 'remove') {
                patches[path] = undefined;
            }
            else if (op.op === 'move') {
                throw new Error('watch move not implemented');
            }
            else if (op.op === 'copy') {
                throw new Error('watch copy not implemented');
            }
            else {
                throw new Error(`watch ${op.op} not implemented`);
            }
        }
        return core_1.size(patches) > 0 ? patches : undefined;
    }
    /**
     * Returns class values of patches.
     */
    getPatchesAndReset() {
        const patches = this.getPatches();
        if (patches) {
            this.reset();
            return patches;
        }
        return;
    }
    /**
     * Returns plain values of patches.
     */
    getPatches() {
        if (this.old) {
            const newState = this.createState();
            const ops = fast_json_patch_1.compare(this.old, newState);
            return this.normalizePatch(ops);
        }
        return;
    }
    /**
     * Returns plain values of patches.
     */
    getRawPatches() {
        const plainPatches = this.getPatches();
        if (plainPatches) {
            return type_1.jsonSerializer.for(Object.getPrototypeOf(this._snapshot).constructor).partialSerialize(plainPatches);
        }
        return;
    }
    /**
     * Applies all detected changes to the original object.
     */
    applyAndReset() {
        const object = this.original;
        const patches = this.getPatches();
        if (patches) {
            for (const [i, v] of core_1.eachPair(patches)) {
                dot_prop_1.set(object, i, v);
            }
        }
        this.reset();
    }
    /**
     * Resets all changes.
     */
    reset() {
        this.old = this.createState();
    }
}
exports.ItemObserver = ItemObserver;
/**
 * Returns patches in plain parameters.
 * @deprecated Use custom patch building.
 */
function observeItem(valueOrSubject, ignore = ['version']) {
    const observer = new ItemObserver(valueOrSubject);
    return observer;
}
exports.observeItem = observeItem;
//# sourceMappingURL=observer.js.map